import unittest
from unittest import TestLoader

if __name__ == '__main__':
    tests = TestLoader().discover('./tests')
    suite = unittest.TestSuite()
    suite.addTests(tests)
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite)



import re

from rework.core.utils import get_root_urls_path


class UrlsHandle:
    """
    Handle django app's urls
    """
    def __init__(self, app=None):
        self.app = app  # When app is None, target to root urls

    @staticmethod
    def _save(f, content):
        f.seek(0)
        f.truncate()
        f.write(content)

    def add_include_urls(self, namespace):
        """
        Add include urls like:

            re_path(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework')),

        """
        if self.app:
            raise NotImplemented

        # Get the root urls file
        urls_file = get_root_urls_path()
        with open(urls_file, 'r+') as f:
            content = f.read()
            content = self._add_route(content, namespace)
            self._save(f, content)

    def _add_route(self, content, namespace):
        rule = f"    re_path(r'^api/{namespace}/', include('rework.contrib.{namespace}.urls')),\n"
        pattern = r'urlpatterns = \[\n([^\n]*\n)*\]\n'
        match = re.search(pattern, content)
        if not match:
            print('There is no `urlpatterns` block in your urls')
            return False

        # get exists rules
        rules = list(match.groups())
        rules.append(rule)

        urlpatterns_block = f'urlpatterns = [\n{"".join(rules)}]\n'

        # make sure import re_path
        if 're_path' not in content:
            urlpatterns_block = f'from django.urls import re_path, include\n\n{urlpatterns_block}'

        content = re.sub(pattern, urlpatterns_block, content)

        return content



"""
Setup server configurations from a new server
Support CentOS 7
"""
from ..hosts import hosts, get_host_value, connect
from ...utils import say, patch_connection_with_say


class SetupServer:
    def __init__(self, c):
        self.c = connect(c)
        self.host_value = get_host_value(c)[1]

        # Patch the `self.c.run`
        self.c = patch_connection_with_say(self.c)

    def __call__(self, *args, **kwargs):
        self.setup_python3()
        self.check_components('nginx') and self.setup_nginx()
        self.check_components('supervisor') and self.setup_supervisor()
        self.check_components('mysql') and self.setup_mysql()
        self.check_components('redis') and self.setup_redis()

    def check_components(self, component):
        return component not in self.host_value.get('exclude_components', [])

    def _remote_exists(self, path):
        exists = self.c.run('[ -e "Python-3.7.9.tgz.1" ] && echo true || echo false')
        return exists == 'true'

    def setup_python3(self):
        """Install python3 and uWSGI"""
        version = '3.7.9'
        major_version = version.rsplit('.', 1)[0]
        # Install Python
        self.c.run('yum -y update')
        self.c.run('yum -y groupinstall "Development tools"')
        try:
            self.c.run('yum -y install wget gcc make zlib-devel')
        except Exception as ex:
            print('ex', ex)

        # Check whether `Python-{version}.tgz` exists
        tgz_file = f'Python-{version}.tgz'
        if not self._remote_exists(tgz_file):
            self.c.run(f'wget https://www.python.org/ftp/python/{version}/{tgz_file}')

        self.c.run(f'tar xzf {tgz_file}')
        self.c.run(
            f'cd Python-{version} && ./configure --with-ssl --prefix=/usr/local && make altinstall'
        )
        self.c.run(f'ln -s /usr/local/bin/python{major_version} /usr/bin/python3')
        self.c.run('python3 -V')
        say('Clean up Python setup files')
        self.c.run(f'rm -rf Python-{version}')

        # Install Gunicorn
        pypi_mirror_suffix = ' -i http://pypi.douban.com/simple/ --trusted-host pypi.douban.com'
        self.c.run(f'python3 -m pip install gunicorn {pypi_mirror_suffix}')

    def setup_nginx(self):
        # Install Nginx
        self.c.run('yum install -y nginx')

    def setup_supervisor(self):
        # Install Supervisor in Python 2
        self.c.run('curl https://bootstrap.pypa.io/pip/2.7/get-pip.py -o get-pip.py -o get-pip.py')
        self.c.run('python get-pip.py')
        self.c.run('python -m pip install supervisor==4.1.0')
        self.c.run('supervisord -c /etc/supervisor/supervisord.conf')  # launch supervisord

    def setup_mysql(self):
        download_url = 'https://dev.mysql.com/get/mysql80-community-release-el7-1.noarch.rpm'
        self.c.run(f'sudo rpm -Uvh {download_url}')
        self.c.run('sudo yum --enablerepo=mysql80-community install mysql-community-server')
        self.c.run('systemctl start mysqld.service')

    def setup_redis(self):
        version = '4.0.14'
        download_url = f'http://download.redis.io/releases/redis-{version}.tar.gz'
        self.c.run(f'curl -o redis-{version}.tar.gz {download_url}')
        self.c.run(f'tar -zxvf redis-{version}.tar.gz')

        self.c.run(f'cd redis-{version} & make')
        self.c.run(f'cd src && make install PREFIX=/usr/local/redis')

        self.c.run(f'mkdir -p /usr/local/redis/conf')
        self.c.run(f'cp redis-{version}/redis.conf /usr/local/redis/conf')

        # start redis
        self.c.run('cd /usr/local/redis && bin/redis-server ./conf/redis.conf')



"""project managements

Using the commands to initialize new project or management existing project
"""

import os
import subprocess

from .handlers.settings import SettingsHandle
from ..utils import say, copy_template_to_file
from ... import core


def init(params):
    """Initialize django rework project"""

    project = params[0]

    base_dir = os.getcwd()
    project_dir = base_dir

    say(f'Initialing project: ``{project}`` using `django-admin` command')
    result = subprocess.run(["django-admin", "startproject", *params])

    if result.returncode != 0:
        say(f'Initialized failed!', icon='🌶 ', wrap='C')
        return False

    # Changed the settings files to satisfy multi environments
    say(f'Changed the settings files to satisfy multi environments')
    settings_folder = os.path.join(project_dir, project)
    settings_handler = SettingsHandle(project=project, path=settings_folder)

    # template variables

    kwargs = {
        'django_rework_version': core.__version__,
        'project': project,
    }

    settings_handler.initialize()

    # fabric DevOps
    copy_template_to_file('fabfile.py', base_dir, **kwargs)

    # Others
    copy_template_to_file('.editorconfig', base_dir, **kwargs)
    copy_template_to_file('.gitignore', base_dir, **kwargs)
    copy_template_to_file('.style.yapf', base_dir, **kwargs)
    copy_template_to_file('requirements.txt', base_dir, **kwargs)

    say(f'Initialized completely!', icon='🎨', wrap='C')



import os
import re
import shutil

from rework import core
from rework.core.utils import copy_template_to_file


class SettingsHandle:
    """
    Handle django settings files
    """

    def __init__(self, project, path):
        self.base_dir = os.getcwd()
        self.project = project
        self.path = path  # settings root path, generally is project name
        self.base_settings_file = None

    def _create_settings_package(self):
        #  1. Make a package named `settings`
        package = os.path.join(self.path, 'settings')
        base_settings_path = os.path.join(package, 'base')
        os.makedirs(base_settings_path)

        #  2. Move origin settings.py to settings/base/__init__.py
        origin_settings_file = os.path.join(self.path, 'settings.py')
        target_settings_file = os.path.join(base_settings_path, '__init__.py')
        shutil.move(origin_settings_file, target_settings_file)

        self.base_settings_file = target_settings_file

    def create_multi_envs(self):
        kwargs = {
            'django_rework_version': core.__version__,
            'project': self.project,
        }
        settings_tpl_path = 'project/settings/'
        envs = ['prod', 'test', 'dev', 'local']
        for env in envs:
            copy_template_to_file(f'{settings_tpl_path}{env}.py', self.base_dir, **kwargs)

    @staticmethod
    def _save(f, content):
        f.seek(0)
        f.truncate()
        f.write(content)

    def _add_tags(self):
        with open(self.base_settings_file, 'r+') as f:
            content = f.read()

            # AUTH_USER_MODEL tag
            content = content.replace(
                '# Password validation', '\n'.join([
                    '# {!AUTH_USER_MODEL}',
                    '',
                    '# Password validation',
                ])
            )
            self._save(f, content)

    def _add_installed_apps(self):
        """Initialize apps pre-installed"""

        with open(self.base_settings_file, 'r+') as f:
            content = f.read()
            content = self._insert_app_to_content(content, 'rest_framework')
            content = self._insert_app_to_content(content, 'rest_framework_simplejwt')
            self._save(f, content)

    @staticmethod
    def _insert_app_to_content(content, app):
        installed_apps_pattern = r'INSTALLED_APPS\s*\=\s*\[[\s\S]*?\]'
        installed_apps_match = re.search(installed_apps_pattern, content)
        if not installed_apps_match:
            print('There is no `INSTALLED_APPS` block in your settings')
            return False

        installed_apps_block = installed_apps_match.group()

        for exist_app in re.findall(r"(?<=').*(?=')", installed_apps_block):
            if exist_app == app:
                return False

        installed_apps_block = re.sub(
            r'\n]',
            f"\n    '{app}',\n]",
            installed_apps_block,
        )

        content = re.sub(installed_apps_pattern, installed_apps_block, content)
        return content

    def _add_rest_framework_setting(self):
        with open(self.base_settings_file, 'r+') as f:
            content = f.read()
            content += """

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework_simplejwt.authentication.JWTAuthentication',
    )
}

"""
            self._save(f, content)

    def initialize(self):
        self._create_settings_package()
        # Added multi env setting files
        self.create_multi_envs()
        # Added template tag to base/__init__.py
        self._add_tags()
        self._add_installed_apps()
        self._add_rest_framework_setting()



import unittest

from rework.core.management.handlers.urls import UrlsHandle


class UrsHandleTestCase(unittest.TestCase):
    def test_add_route_to_urlpatterns(self):
        content = """

from django.contrib import admin
from django.urls import path

urlpatterns = [
    path('admin/', admin.site.urls),
]

"""
        except_content = """

from django.contrib import admin
from django.urls import path

from django.urls import re_path, include

urlpatterns = [
    path('admin/', admin.site.urls),
    re_path(r'^api/tests/', include('rework.contrib.tests.urls', namespace='tests')),
]

"""
        urls_handler = UrlsHandle()
        content = urls_handler._add_route(content, 'tests')
        self.assertEqual(content, except_content)



"""
Deploy to server 
"""
import os

from ..hosts import get_host_value, connect
from ... import devops
from ...utils import say


class Deploy:
    def __init__(self, c):
        self.c = connect(c)
        self.host_value = get_host_value(c)[1]

        self.project = devops.PROJECT
        self.env = devops.ENV

        self.host = self.host_value.get('host')
        self.port = self.host_value.get('port', 22)
        self.password = self.host_value.get('password')

        # Should reload nginx when it's configurations file changed
        self._should_reload_nginx = False

    def __call__(self, *args, **kwargs):
        # supervisor service name
        service = f'{self.project}_{self.env}'
        if self.env == 'prod':
            env_root = f'/opt/projects/'
            branch = 'master'
        elif self.env == 'test':
            env_root = f'/opt/test-projects/'
            branch = 'test'
        elif self.env == 'dev':
            env_root = f'/opt/dev-projects/'
            branch = 'dev'
        else:
            raise Exception('Un-supported Environment to deploy.')
        root = f'{env_root}{self.project}-server/'

        say(f'Project root is: {root}')

        self.pull(branch, root)

        if kwargs.get('infrastructure'):
            self.setup_infrastructure(root)

        if kwargs.get('requirements_update'):
            self.update_requirement(root)

        self.migrate(root)
        self.collect_static(root)
        self.restart(service)

    def _copy_nginx(self, root):
        """Copy nginx files"""
        nginx_path = f'{root}.deploy/nginx/'

        origin = f'{nginx_path}{self.project}_{self.env}.conf'

        destination = f'/etc/nginx/conf.d/'
        self.c.run(f'cp {origin} {destination}')
        self._should_reload_nginx = True
        say('Copied nginx configuration successfully')

    def _copy_supervisor(self, root):
        """Copy supervisor files

        Compatible with Django-Rework ~0.2:
            supervisor file in django-rework <= 0.2:
                .deploy/supervisor/{self.project}_supervisor_{self.env}.conf
            supervisor file in django-rework >= 0.3:
                .deploy/supervisor/{self.project}_{self.env}.conf
        """
        supervisor_path = f'.deploy/supervisor/'

        origin = f'{supervisor_path}{self.project}_{self.env}.conf'
        if not os.path.exists(origin):
            say(f'Supervisor file: {origin} not exists, try find another...')
            origin = f'{supervisor_path}{self.project}_supervisor_{self.env}.conf'

        origin = f'{root}{origin}'

        destination = f'/etc/supervisor/conf.d/'
        self.c.run(f'cp {origin} {destination}')
        say('Copied supervisor configuration successfully')

    def pull(self, branch, root):
        # Pull latest code
        say('Pull latest code from remote git')
        self.c.run(f'cd {root} && git checkout {branch} && git pull')

    def setup_infrastructure(self, root):
        say('Setup infrastructure')
        self._copy_supervisor(root)
        self._copy_nginx(root)

    def update_requirement(self, root):
        # Update requirements
        say('Install requirements')
        self.c.run(f'cd {root} && python3 -m pip install -r requirements.txt')

    def migrate(self, root):
        # Models migrate
        say('Migrate database')
        self.c.run(
            f'cd {root} && python3 manage.py migrate --settings={self.project}.settings.{self.env}'
        )

    def collect_static(self, root):
        # Collect static
        say('Collect static')
        settings_suffix = f'--settings={self.project}.settings.{self.env}'
        self.c.run(f'cd {root} && python3 manage.py collectstatic {settings_suffix} --no-input')

    def restart(self, service):
        # Restart infrastructure
        say('Restart Supervisor and Nginx(if conf changed)')
        self.c.run(f'supervisorctl restart {service}')
        if self._should_reload_nginx:
            self.c.run(f'nginx -s reload')



import os

from ..utils import say, copy_template_to_file, get_project_name
from ..utils.command import BaseCommand
from ... import core


class DeployCommand(BaseCommand):
    def handle(self, params):
        if params == ['--init']:
            # template variables
            project = get_project_name()
            kwargs = {
                'django_rework_version': core.__version__,
                'project': project,
            }
            deploy_path = os.path.join(self.base_dir, 'deploy')

            copy_queues = [
                ('.deploy/nginx/project_prod.conf', deploy_path),
                ('.deploy/nginx/project_test.conf', deploy_path),
                ('.deploy/supervisor/supervisord.conf', deploy_path),
                ('.deploy/supervisor/project_prod.conf', deploy_path),
                ('.deploy/supervisor/project_test.conf', deploy_path),
            ]
            for q in copy_queues:
                copy_template_to_file(q[0], self.base_dir, **kwargs)

            say('Deploy files copied successfully')

        else:
            say('Unknown command args!')

    def __call__(self, params):
        return self.handle(params)



from django.urls import path
from rest_framework_simplejwt.views import (
    TokenObtainPairView,
    TokenRefreshView,
)

urlpatterns = [
    path('auth/token', TokenObtainPairView.as_view(), name='token_obtain_pair'),
    path('auth/token/refresh', TokenRefreshView.as_view(), name='token_refresh'),
]
